# See article: http://fiber-space.de/wordpress/?p=1580
# for a more detailed description.

__all__ = ["PrefixTree", "fuzzy_match"]

class PrefixTree(object):
    '''
    Tree used to store a wordlist. Words are factored according to their common prefixes.

    For performance reasons the PrefixTree is only partially build and will be further constructed
    by need.
    '''
    def __init__(self, char = '', parent = None):
        self.char      = char
        self.parent    = parent
        self.is_word   = False
        self._children = {}
        self._words    = set()

    def _get_children(self):
        if self._words:
            self._create_children()
        return self._children

    children = property(_get_children)

    def _create_children(self):
        for tree, word in self._words:
            tree.insert(word)
        self._words = set()

    def _tolist(self):
        if self.is_word:
            yield self.trace()
        for p in self.children.values():
            for s in p._tolist():
                yield s

    def __iter__(self):
        return self._tolist()

    def insert(self, value):
        if value:
            c = value[0]
            tree = self._children.get(c)
            if tree is None:
                tree = PrefixTree(c, self)
                self._children[c] = tree
            if len(value) == 1:
                tree.is_word = True
            tree._words.add((tree,value[1:]))
        else:
            self.is_word = True

    def __contains__(self, value):
        if value:
            if value in self._words:
                return True
            c = value[0]
            if c in self._children:
                return value[1:] in self._children[c]
            return False
        return True

    def __len__(self):
        if self.parent is not None:
            return len(self.parent)+1
        return 0

    def trace(self):
        if self.parent is not None:
            return self.parent.trace()+self.char
        return self.char

def update_visited(ptree, visited):
    visited[ptree][-1] = 0
    T = ptree.parent
    while T is not None and T.char!='':
        if len(T.children) == 1:
            visited[T][-1] = 0
            T = T.parent
        else:
            return

def is_visited(i, T, k, visited):
    d = visited.get(T, {})
    if -1 in d:
        return True
    m = d.get(i,-1)
    if k>m:
        d[i] = k
        visited[T] = d
        return False
    return True


def fuzzy_match(S, ptree, k, i=0, visited = None, N = 0):
    '''
    Searches for strings in a PrefixTree ( wordlist ) with a bound number of errors.

    More precisely: computes all strings T contained in ptree with a distance dist(S, T)<=k.
    '''
    trees = set()
    # handles root node of a PrefixTree
    if ptree.char == '' and ptree.children:
        N = len(S)
        S+='\0'*(k+1)
        visited = {}
        for T in ptree.children.values():
            trees.update(fuzzy_match(S, T, k, i, visited, N))
        return trees

    # already tried
    if is_visited(i, ptree, k, visited):
        return []

    # can't match ...
    if k == -1 or (k == 0 and S[i] != ptree.char):
        return []

    if ptree.is_word and (N-i<=k or (N-i<=k+1 and ptree.char == S[i])):
        trees.add(ptree.trace())
        if not ptree.children:
            update_visited(ptree, visited)
            return trees

    if ptree.char!=S[i]:
        trees.update(fuzzy_match(S, ptree, k-1, i+1, visited, N))

    for c in ptree.children:
        if ptree.char == S[i]:
            trees.update(fuzzy_match(S, ptree.children[c], k, i+1, visited, N))
        else:
            trees.update(fuzzy_match(S, ptree.children[c], k-1, i+1, visited, N))
        trees.update(fuzzy_match(S, ptree.children[c], k-1, i, visited, N))
    return trees


