from __future__ import with_statement  # Python 2.5 compatible
import string
import time

from fuzzy_compare import*
from fuzzy_match import*

wordlist = set(line.strip().lower() for line in open("german.txt").readlines())

class timing(object):
    def __enter__(self):
        import time
        self.start = time.time()
        return self

    def __exit__(self, *args):
        self.end = time.time()
        self.delta = self.end - self.start



def test_fuzzy_match():
    '''
    For a given ( small ) wordlist, represented as a PrefixTree, this function checks that for
    each word w:

        for k in 0..len(w) the following must hold:

            for S in fuzzy_compare(w, wordlist,k):

               fuzzy_compare(w, S, k) == True

            for S not in fuzzy_compare(w, wordlist,k) then

               fuzzy_compare(w, S, k) == False

    '''
    words = ['baum',
             'graus',
             'haus',
             'hals',
             'halle',
             'kralle',
             'mau',
             'maus',
             'mauser',
             'mausern',
             'mann',
             'wanne']

    pt = PrefixTree()
    for word in words:
        pt.insert(word)

    def subtest():
        for t in words:
            for k in range(len(t)):
                L = fuzzy_match(t, pt, k)
                for item in L:
                    assert fuzzy_compare(t, item, k) == True, (t, item, k, L)
                for item in pt:
                    if item not in L:
                        assert fuzzy_compare(t, item, k) == False, (t, item, k, L)
    subtest()

def levenshtein(s1, s2):
    l1 = len(s1)
    l2 = len(s2)
    matrix = [range(l1 + 1)] * (l2 + 1)
    for zz in range(l2 + 1):
      matrix[zz] = range(zz,zz + l1 + 1)
    for zz in range(0,l2):
      for sz in range(0,l1):
        if s1[sz] == s2[zz]:
          matrix[zz+1][sz+1] = min(matrix[zz+1][sz] + 1, matrix[zz][sz+1] + 1, matrix[zz][sz])
        else:
          matrix[zz+1][sz+1] = min(matrix[zz+1][sz] + 1, matrix[zz][sz+1] + 1, matrix[zz][sz] + 1)
    return matrix[l2][l1]

def string_set(S = None, k = 0, strings = None, n = 6):
    if S is None:
        strings = []
        S = ["a"]*n
        strings.append("".join(S))
    for i in range(k, n):
        S1 = S[:]
        S1[i] = "b"
        strings.append("".join(S1))
        string_set(S1, i+1, strings, n)
    return strings

def string_pairs(n):
    L1 = string_set(n=n)
    pairs = []
    for i in range(len(L1)):
        for k in range(1, n+1):
            L2 = string_set(n=k)
            for j in range(len(L2)):
                pairs.append((L1[i],L2[j],levenshtein(L1[i], L2[j])))
                pairs.append((L2[j],L1[i],levenshtein(L2[j], L1[i])))
    return pairs

def string_pair(k=15):
    S  = "".join(str(s) for s in range(10)*k)
    S2 = "".join(str(s) for s in [0 if s%3==0 else s for s in range(10)*k])
    return S, S2, levenshtein(S, S2)


def test_fuzzy_compare(n):
    L = string_pairs(n)
    for s1, s2, d in L:
        assert fuzzy_compare(s1, s2, d)   == True, (s1, s2, d)
        assert fuzzy_compare(s1, s2, d-1) == False, (s1, s2, d-1)



def timings():
    with timing() as T:
        pt = PrefixTree()
        for word in wordlist:
            pt.insert(word)

    print "Build (partial) PrefixTree in time:", T.delta

    with timing() as T:
        fuzzy_match("haus", pt, 1)

    print "1st match of 'haus' (k = 1):", T.delta

    with timing() as T:
        fuzzy_match("haus", pt, 1)

    print "2nd match of 'haus' (k = 1):", T.delta

    with timing() as T:
        fuzzy_match("wunsch", pt, 2)

    print "1st match of 'wunsch' (k = 2):", T.delta

    with timing() as T:
        fuzzy_match("wunsch", pt, 2)

    print "2nd match of 'wunsch' (k = 2):", T.delta

    with timing() as T:
        fuzzy_match("meise", pt, 3)

    print "1st match of 'meise' (k = 3):", T.delta

    with timing() as T:
        fuzzy_match("meise", pt, 3)

    print "2nd match of 'meise' (k = 3):", T.delta



if __name__ == '__main__':
    test_fuzzy_match()
    test_fuzzy_compare(6)
    timings()

